require "skirental_api/railtie"

require "faraday"
require "faraday_middleware"
require "faraday-http-cache"

require "symbolize_helper"

require 'uri'
require 'net/http'

class Faraday::AddSrTokens < Faraday::Middleware
    def initialize(app, options = {})
        super(app)
    end
    def call(env)
        config = Rails.application.config.services[:SR_API]
        sr_ticket = getTicket(config)
        sr_hash= Digest::SHA1.hexdigest("SPORTSRENTAL" + sr_ticket + config[:login])
        env[:request_headers]["SrTicket"] = sr_ticket
        env[:request_headers]["SrHash"]   = sr_hash
        @app.call(env)
    end

    def getTicket(config)
        # First get a valid ticket
        current_ticket = SkirentalApi::Client.sr_api_token

        if current_ticket.present?
            expires = current_ticket["expires"]
            seconds_left = Time.parse(expires) - Time.now
            puts "ticket seconds left: #{seconds_left}"
            # dispose the ticket if it is invalid (with some buffer for security)
            current_ticket = nil if seconds_left < 2
        end

        if current_ticket.present?
            return current_ticket["ticket"]
        else
            service_url = config[:base_url]

            url = URI("#{service_url}/ticket")

            http = Net::HTTP.new(url.host, url.port)
            http.use_ssl = (url.scheme == 'https')
            request = Net::HTTP::Get.new(url)
            request.basic_auth config[:login], config[:password]
            request["Cache-Control"] = 'no-cache'
            response = http.request(request)

            SkirentalApi::Client.sr_api_token = JSON.parse(response.read_body)
            return SkirentalApi::Client.sr_api_token["ticket"]
        end
    end
end

# Hash cleaner
class Hash
    def clean!
        self.delete_if do |key, val|
            if block_given?
                yield(key,val)
            else
                # Prepare the tests
                test1 = val.nil?
                test2 = val === 0
                test3 = val === false
                test4 = val.empty? if val.respond_to?('empty?')
                test5 = val.strip.empty? if val.is_a?(String) && val.respond_to?('empty?')

                # Were any of the tests true
                test1 || test2 || test3 || test4 || test5
            end
        end

        self.each do |key, val|
            if self[key].is_a?(Hash) && self[key].respond_to?('clean!')
                if block_given?
                    self[key] = self[key].clean!(&Proc.new)
                else
                    self[key] = self[key].clean!
                end
            end
        end

        return self
    end
end

module SkirentalApi
    class Client
        using SymbolizeHelper

        mattr_accessor :sr_api_token

        ACTION_STATE_INIT       = 0
        ACTION_STATE_CREATED    = 1
        ACTION_STATE_UPDATED    = 2
        ACTION_STATE_PAID       = 4
        ACTION_STATE_CONFIRMED  = 8
        ACTION_STATE_CANCELED   = 16
        ACTION_STATE_MAILED     = 32
        ACTION_STATE_FETCHED    = 64
        ACTION_STATE_FINISHED   = 128
        ACTION_STATE_READY      = 256
        ACTION_STATE_SHOPMAILSENT  = 512
        ACTION_STATE_READYMAILSENT = 1024
        ACTION_STATE_CHANGED    = 2048
        ACTION_STATE_MAILREAD   = 4096
        ACTION_STATE_GHOST      = 8192

        ACTION_GHOST            = -1 # Durch den Aktionstyp  'Ghost' wird der Buchungstatus mit 'Ghost' verodert, um die Buchung als vorläufig zu makieren. Die Buchung wird nur für einige Minuten vorgehalten und anschliessend nicht mehr ausgewertet/gelöscht.
        ACTION_GET              = 0
        ACTION_CREATE           = 1
        ACTION_UPDATE           = 2
        ACTION_CONFIRM          = 4
        ACTION_CANCEL           = 5
        ACTION_ACCEPT           = 6


        def self.sync_booking_state(booking)
            return nil if booking.context[:customers].blank?
            current_locale = booking.context[:locale] || "de"
            current_locale = current_locale.split("-").first

            booking_data = nil
            shop_lic = self.get_mapping[booking.shop_id.downcase]
            fail "shop_lic was not found" if shop_lic.nil?

            if booking.booking_id.present? # The content is already created.

                last_payment = booking.last_payment
                if last_payment.present? && (last_payment.is_payed? || last_payment.is_reserved?) # payment/reservation alreay finished
                    if booking.is_checkin
                        puts "sync_booking_state with ACTION_CONFIRM in checkin mode"
                        booking_data = self.create_booking(booking, ACTION_CONFIRM) # allow update of everything in checkin mode
                        self.insert_personal_data(booking, booking_data)
                    else
                        # get Booking data from SR API
                        booking_data = self.get_booking(booking.booking_id)
                        fail "No Booking Context available with booking id #{booking.booking_id}" if booking_data.empty?
                        
                        puts "sync_booking_state with ACTION_CONFIRM"
                        self.insert_personal_data(booking, booking_data) # only allow update of personal data, not products.
                        booking_data[:action] = ACTION_CONFIRM # updates for already finished reservations need action == 4
                    end
                else
                    if booking.context[:isShortRent]
                        puts "sync_booking_state with ACTION_GHOST"
                        booking_data = self.create_booking(booking, ACTION_GHOST) # Ghost (Pre-Reserve item availabilites)
                    else
                        puts "sync_booking_state with ACTION_UPDATE"
                        booking_data = self.create_booking(booking, ACTION_UPDATE) # Update
                    end
                    self.insert_personal_data(booking, booking_data)
                end
            else
                # Only create a booking code via creating an empty context

                # First create an empty booking context and ..
                puts "sync_booking_state with ACTION_CREATE"
                booking_data = self.booking_data_structure(shop_lic, booking.creation_time)
                booking_data[:action] = ACTION_CREATE
            end

            base_url = Rails.application.config.frontend_api[:base_url]
            booking_data[:frontendURL] = "#{base_url}/#{current_locale}/#{booking.shop_id}/reservation?session=#{booking.session_id}&readmail=1"

            # Send it to the S|R-Api
            response = self.post_booking(shop_lic, booking_data.clean!, current_locale)

            # Return the created booking code.
            if response.blank? || response.deep_symbolize_keys[:code].blank?
                return nil
            else
                return response.deep_symbolize_keys
            end
        end

        def self.confirm_booking(booking, is_reservation)
            return nil if booking.context[:customers].blank?
            current_locale = booking.context[:locale] || "de"
            current_locale = current_locale.split("-").first

            fail "No Booking ID available! Missing call to sync_booking_state?" unless booking.booking_id.present?

            puts "confirm_booking…"
            booking_data = self.create_booking(booking, ACTION_CONFIRM)
            self.insert_personal_data(booking, booking_data)

            if not is_reservation
                # Status = Oder-Verknüpfung mit paid, wenn die Zahlung erfolgt ist
                last_payment = booking.last_payment
                fail "No payment was found for this booking!" if last_payment.nil?

                booking_data[:client][:card] = {
                    id:             0,
                    status:         "Complete",
                    tid:            last_payment.payment_reference,
                    provider:       last_payment.provider_name,
                    cur:            last_payment.transaction[:amount][:currency],
                    price:          last_payment.transaction[:amount][:total],
                }
                booking_data[:state] |= ACTION_STATE_PAID
            end

            booking_data[:state] |= ACTION_STATE_CONFIRMED

            shop_lic = self.get_mapping[booking.shop_id.downcase]
            fail "shop_lic was not found" if shop_lic.nil?

            # .. send it to the S|R-Api
            response = self.post_booking(shop_lic, booking_data.clean!, current_locale)

            # Return the created voucher code
            if response.blank? || response.deep_symbolize_keys[:code].blank?
                return nil
            else
                return response.deep_symbolize_keys
            end
        end

        # Returns updated booking_data
        def self.insert_personal_data(booking, booking_data)
            # upsert customer notice
            booking_data[:note] = booking.context[:customerNotice]

            # Update personal data of main customer
            main_customer_personal_data = booking[:context][:customers].first[:personalData]
            self.set_customer_data(booking_data[:client], main_customer_personal_data)

            # Now update all other personal data for the siblings. Skip first, as it is the main customer
            ( 1 ... (booking[:context][:customers].count) ).each do |customer_index|
                # Skip setting customer-data until there is any booking data created. This will happen in confirm_booking..
                next if booking_data[:bookings][customer_index - 1].nil?
                self.set_customer_data(booking_data[:bookings][customer_index - 1][:client], booking[:context][:customers][customer_index][:personalData])
            end

            # Update selected variants of main customer
            self.set_customer_selected_package_variants(booking_data[:orders], booking[:context][:customers].first[:packages])
            self.set_customer_selected_product_variants(booking_data[:orders], booking[:context][:customers].first[:products])

            # Now update selected product variants
            ( 1 ... (booking[:context][:customers].count) ).each do |customer_index|
                # Skip setting customer-data until there is any booking data created. This will happen in confirm_booking..
                next if booking_data[:bookings][customer_index - 1].nil?
                self.set_customer_selected_package_variants(booking_data[:bookings][customer_index - 1][:orders], booking[:context][:customers][customer_index][:packages])
                self.set_customer_selected_product_variants(booking_data[:bookings][customer_index - 1][:orders], booking[:context][:customers][customer_index][:products])
            end
        end

        def self.paypal_credentials(shop_id)
            payment_systems = self.settings(shop_id)[:money][:payment][:ipMs]

            payment_data = payment_systems.select { |ipms| ipms[:type] == "paypal" }.first

            return nil if payment_data.nil?

            {
                mode:           "live",
                client_id:      payment_data[:id],
                client_secret:  payment_data[:secret]
            }
        end

        def self.customer_groups(shop_id, shop_lang)
            all_price_groups = self.price_info(shop_id, shop_lang)

            customer_groups = []
            all_price_groups.each do |price_group|
                customer_group = {
                    id:             price_group[:iDs].first.to_s,
                    name:           price_group[:alias].blank? ? price_group[:name] : price_group[:alias],
                    ageFrom:        price_group[:ageFrom],
                    ageTo:          price_group[:ageTo],
                    canOrder:       price_group[:canOrder]
                }
                customer_groups.push customer_group
            end

            customer_groups.sort_by { |group| group[:ageFrom] }
        end


        def self.valid_booking_range(shop_settings)
            dates = shop_settings[:dates]
            valid_range = dates[:valid]
            lead_range  = dates[:lead]

            start_date = nil
            end_date = nil
            max_number_of_days = nil
            not_before_min_days = nil
            not_bofore_max_days = nil
            today_ok_time = nil

            if valid_range[:maxLength] > 0
                max_number_of_days = valid_range[:maxLength]
            end

            # die  datumsbereiche sind nur gültig, wenn alle 4 werte gesetzt sind. Sonst ist alles gültig
            if (valid_range[:monthFrom] > 0  && valid_range[:dayFrom] > 0  && valid_range[:monthUntil] > 0 && valid_range[:dayUntil] > 0)
                start_date = Date.new(Date.current.year, valid_range[:monthFrom], valid_range[:dayFrom])
                end_date = Date.new(Date.current.year, valid_range[:monthUntil], valid_range[:dayUntil])
                
                if start_date > end_date
                    end_date = Date.new(Date.current.year + 1, valid_range[:monthUntil], valid_range[:dayUntil])
                end
                if Date.current > end_date
                    start_date = Date.new(Date.current.year + 1, valid_range[:monthFrom], valid_range[:dayFrom])
                    end_date = Date.new(Date.current.year + 1, valid_range[:monthUntil], valid_range[:dayUntil])
                end
            else
                start_date = Date.current
                end_date   = Date.current + 365.days # TODO: review with Keller
            end

            if lead_range[:todayOkTime].present? && lead_range[:todayOkTime] > 0
                today_ok_time = lead_range[:todayOkTime]
            end

            if lead_range[:min] != 0
                not_before_min_days = lead_range[:min]
            end

            if lead_range[:max] != 0
                not_before_max_days = lead_range[:max]
            end

            {
                firstValidDay: start_date,
                lastValidDay: end_date,
                lastValidBookingTimeForToday: today_ok_time,
                maxBookingDuration: max_number_of_days,
                minDaysBeforeBooking: not_before_min_days,
                maxDaysBeforeBooking: not_before_max_days,
                blockedDays: [], # Not supported
                hideStartDate: dates[:hideStart].present? ? dates[:hideStart] : false,
                hideEndDate: dates[:hideUntil].present? ? dates[:hideUntil] : false,
                arrivalMaxDaysBeforeStart: lead_range[:maxDaysArival] # TODO: (sic!)
            }
        end

        def self.end_of_season(shop_settings)
            valid_range = shop_settings[:dates][:valid]
            season = shop_settings[:dates][:season]
            # die datumsbereiche sind nur gültig, wenn alle 4 werte gesetzt sind. Sonst ist alles gültig
            if (valid_range[:monthFrom] > 0  && valid_range[:dayFrom] > 0  && valid_range[:monthUntil] > 0 && valid_range[:dayUntil] > 0 && season[:monthend] > 0 && season[:dayend] > 0 )
                start_date = Date.new( Date.current.year, valid_range[:monthFrom], valid_range[:dayFrom] )
                if valid_range[:monthUntil] < valid_range[:monthFrom]
                    end_date = Date.new( Date.current.year + 1, season[:monthend], season[:dayend] )
                else
                    end_date = Date.new( Date.current.year, season[:monthend], season[:dayend] )
                end
            else
                start_date = Date.current
                end_date   = Date.current + 120.days
            end

            end_date
        end

        def self.shop_datum(shop_id, shop_lang)
            shop_data = self.shop(shop_id, shop_lang)
            shop_settings = self.settings(shop_id, shop_lang)
            children_discount = {}
            unless shop_settings[:money][:discount][:kidLimit].nil?
                children_discount = {
                    percentage: shop_settings[:money][:discount][:kidLimit][:value],
                    maxNumberOfChildren: shop_settings[:money][:discount][:kidLimit][:maxkids],
                    maxAgeForChildren: shop_settings[:money][:discount][:kidLimit][:maxage],
                    minNumberOfAdults: shop_settings[:money][:discount][:kidLimit][:minadult]
                }
            end
            early_birds_discounts = []
            unless shop_settings[:money][:discount][:earlyBird].nil?
                shop_settings[:money][:discount][:earlyBird].each do |early_birds_discount|
                    early_birds_discounts.push({
                        name: early_birds_discount[:name],
                        percentage: early_birds_discounnt[:value],
                        daysBeforeBooking: early_birds_discount[:days]
                    })
                end
            end
            payment_providers = []
            shop_settings[:money][:payment][:ipMs].each do |payment_provider|
                payment_providers.push({
                    id: payment_provider[:id],
                    type: payment_provider[:type],
                    discount: payment_provider[:dsc],
                    show: payment_provider[:show],
                    secret: payment_provider[:secret],
                    fee: payment_provider[:fee]
                })
            end

            customer_groups = self.customer_groups(shop_id, shop_lang)

            categorySpecificMandatoryFields = {}
            shop_settings[:mandatory][:typeDataFields].each do |mandatoryFields|
                categorySpecificMandatoryFields[mandatoryFields[:type]] = {
                    :birthday          => mandatoryFields[:birthday],
                    :height            => mandatoryFields[:size],
                    :weight            => mandatoryFields[:weight],
                    :headCircumference => mandatoryFields[:head],
                    :clothingSize      => mandatoryFields[:confection],
                    :shoeSize          => mandatoryFields[:shoesize],
                    :shoeSizeSystem    => mandatoryFields[:shoetype],
                    :skiSkillLevel     => mandatoryFields[:ski],
                    :snowboardPosition => mandatoryFields[:snobo]
                }
            end


            css_data = shop_data[:shopInfo][:css]

            valid_booking_range = self.valid_booking_range(shop_settings)
            end_of_season       = self.end_of_season(shop_settings)

            google_maps_config  = Rails.application.config.services[:GOOGLE_MAPS]
            google_maps_api_key = nil
            google_maps_api_key = google_maps_config[:googleMapsAPIKey] if google_maps_config.present?

            config = Rails.application.config.services[:SR_API]

            shop_options = shop_settings[:options]

            # Finally create the returned hash
            {
                id: shop_id,
                slug: shop_data[:slug],
                lic: shop_data[:lic],
                masterLic: shop_data[:masterLic],
                name: shop_data[:name],
                title: shop_data[:shopInfo][:title],
                subtitle: shop_data[:shopInfo][:title2],
                descriptions: [
                    shop_data[:shopInfo][:info],
                    shop_data[:shopInfo][:info2]
                ],
                texts: {
                    confirmationText: shop_data[:shopInfo][:confirmationText].present? ? shop_data[:shopInfo][:confirmationText] : nil,
                    reserveButtonLabel: shop_data[:shopInfo][:reserveButtonLabel].present? ? shop_data[:shopInfo][:reserveButtonLabel] : nil,
                },
                favicon: shop_data[:shopInfo][:favIcon].present? ? shop_data[:shopInfo][:favIcon] : nil,
                mapIcon: shop_data[:shopInfo][:mapIcon].present? ? shop_data[:shopInfo][:mapIcon] : nil,
                logo: shop_data[:shopInfo][:logo].present? ? shop_data[:shopInfo][:logo] : nil,
                logoLink: shop_data[:shopInfo][:logoLink],
                homeLink: shop_data[:shopInfo][:homeLink],
                images:  shop_data[:shopInfo][:img].split("\;").map { |e| "https://www.sports-rental.info/banner.aspx?shop=#{shop_data[:lic]}&pic=#{e}" },
                openingTimes: [
                    shop_data[:shopInfo][:open1],
                    shop_data[:shopInfo][:open2]
                ],
                address: shop_data[:address],
                addressLines: [
                    shop_data[:shopInfo][:addline1],
                    shop_data[:shopInfo][:addline2],
                    shop_data[:shopInfo][:addline3],
                    shop_data[:shopInfo][:addline4]
                ],
                coordinates: {
                    lat: shop_data[:location][:lat],
                    lng: shop_data[:location][:lng]
                },
                branches:               shop_data[:subShops],
                gtc:                    shop_data[:notes][:gtc],
                gtcUrl:                 shop_data[:notes][:gtcUrl],
                privacy:                shop_data[:notes][:privacy],
                privacyUrl:             shop_data[:notes][:privacyUrl],
                imprint:                shop_data[:notes][:imprint],
                imprintUrl:             shop_data[:notes][:imprintUrl],
                googleMapsAPIKey:       google_maps_api_key || "AIzaSyBgiBqNNvbEgL_DgQ7no9CH_7sVsd80jCE" ,
                checkoutInfo:           shop_data[:notes][:addConf],
                highlights:             shop_data[:highlights],
                timezone:               shop_data[:location][:timezone],
                utcOffset:              shop_data[:location][:gmt],
                dateInfo:               shop_data[:location][:dateinfo],
                skipCart:               shop_settings[:skipCart],
                hideShopInfo:           shop_settings[:hideShopInfo].present? ? shop_settings[:hideShopInfo] : false,
                hideCookieBanner:       shop_settings[:hideCookieBanner].present? ? shop_settings[:hideCookieBanner] : false,
                defaultShoeSizeSystem:  shop_settings[:defaults][:defshoe],
                defaultLanguage:        shop_settings[:defaults][:deflang],
                currentLanguage:        shop_data[:lang],
                availableLanguageCodes: shop_data[:languages],
                availableLanguages:     shop_data[:lang2Countries].present? ? shop_data[:lang2Countries] : shop_data[:languages],
                currencyCode:           shop_settings[:money][:payment][:cur],
                currencySymbol:         "#{shop_settings[:money][:cur]} ", # NOTE: space after symbol for display purposes in angular frontend!
                isBookingEditable:      shop_settings[:editBooking],
                isPaymentOptional:      shop_settings[:money][:payment][:optional],
                paymentProviders:       payment_providers,
                vat:                    shop_settings[:money][:tax][:value],
                showVat:                shop_settings[:money][:tax][:show],
                includeVat:             shop_settings[:money][:tax][:include],
                insuranceInfo:          shop_data[:notes][:insurance],
                insurance:              shop_settings[:money][:insurance][:value],
                showInsurance:          shop_settings[:money][:insurance][:show],
                isInsuranceOptional:    shop_settings[:money][:insurance][:optional],
                onlineDiscount:         shop_settings[:money][:discount][:value],
                deposit:                shop_settings[:money][:deposit][:value],
                depositType:            shop_settings[:money][:deposit][:type],
                childrenDiscount:       children_discount,
                earlyBirdDiscounts:     early_birds_discounts,
                validBookingRange:      valid_booking_range,
                seasonalRental:         {
                    active:          shop_settings[:dates][:season][:use],
                    endOfSeason:     end_of_season,
                    defaultDuration: shop_settings[:dates][:season][:lengthRental]
                },
                shortRent:              shop_settings[:shortRent],
                marketplace:            shop_data[:marketplace],
                checkin:                shop_data[:checkin].present? ? shop_data[:checkin] : false,
                checkinProductsOptional: shop_settings[:productsOpt].present? ? shop_settings[:productsOpt] : false,
                maxClients:             shop_settings[:maxClients],
                customerGroups:         customer_groups,

                # Mandatory Data
                globalMandatoryPersonalDataFields: shop_settings[:mandatory][:clientFields],
                globalOptionalPersonalDataFields: shop_options.merge(shop_settings[:mandatory][:clientBaseFields]),
                categorySpecificMandatoryFields:  categorySpecificMandatoryFields,

                # CSS-Data
                themeColors: {
                    primary:    css_data[:primaryColor].present?    ? css_data[:primaryColor]   : "#0042A3",
                    selection:  css_data[:selectionColor].present?  ? css_data[:selectionColor] : "#61A90D",
                    highlight:  css_data[:highlightColor].present?  ? css_data[:highlightColor] : "#FF012F",
                    text:       css_data[:textColor].present?       ? css_data[:textColor]      : "#4A4A4A",
                    stars0:     css_data[:stars0Color].present?     ? css_data[:stars0Color]    : "#F7F7F7",
                    stars1:     css_data[:stars1Color].present?     ? css_data[:stars1Color]    : "#4AE2C4",
                    stars2:     css_data[:stars2Color].present?     ? css_data[:stars2Color]    : "#F5A623",
                    stars3:     css_data[:stars3Color].present?     ? css_data[:stars3Color]    : "#E4396D",
                    stars4:     css_data[:stars4Color].present?     ? css_data[:stars4Color]    : "#4A90E2",
                    stars5:     css_data[:stars5Color].present?     ? css_data[:stars5Color]    : "#4A4A4A",
                    stars0text: css_data[:stars0TxtColor].present?  ? css_data[:stars0TxtColor] : "#4A4A4A",
                    stars1text: css_data[:stars1TxtColor].present?  ? css_data[:stars1TxtColor] : "#FFFFFF",
                    stars2text: css_data[:stars2TxtColor].present?  ? css_data[:stars2TxtColor] : "#FFFFFF",
                    stars3text: css_data[:stars3TxtColor].present?  ? css_data[:stars3TxtColor] : "#FFFFFF",
                    stars4text: css_data[:stars4TxtColor].present?  ? css_data[:stars4TxtColor] : "#FFFFFF",
                    stars5text: css_data[:stars5TxtColor].present?  ? css_data[:stars5TxtColor] : "#FFFFFF"
                },
                themeFont: {
                    family: css_data[:fontFamily].present? ? css_data[:fontFamily] : nil,
                    url: css_data[:fontUrl].present? ? css_data[:fontUrl] : nil
                },
                sliderSeparator: {
                    color: css_data[:sliderTxtColor].present? ? css_data[:sliderTxtColor] : nil,
                    backgroundImageUrl: css_data[:sliderBgUrl].present? ? css_data[:sliderBgUrl] : nil
                },
                imageBaseUrl: config[:image_base_url]
            }
        end

        def self.categories(shop_id, shop_lang)
            categories = self.sr_products(shop_id, shop_lang)
            return_categories = []
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank? && category[:packages].blank?

                ret_hash = {
                    id: category[:id].to_s,
                    name: category[:name],
                    type: category[:type],
                    packages: {},
                    products: {}
                }
                return_categories.push ret_hash
            end

            return_categories
        end

        def self.packages(shop_id, product_category_id, start_date, duration, customer_group_id, shop_lang)
            age = self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            categories = self.sr_products(shop_id, shop_lang)
            return_packages = []
            categories.each do |category|
                next if category[:packages].blank?

                # This lookup is kind of slow. We might need to improve later, if necessary
                next if category[:id].to_s != product_category_id

                category[:packages].each do |package|
                    valid_prices = package[:prices].select { |p| age >= p[:ageFrom] && age <= p[:ageTo] }
                    next if valid_prices.empty?

                    # Main Product ist das erste nicht optionale Element in der Liste der items
                    main_product = package[:items].select { |item| item[:optional] == false }.first

                    # Accessories are all products which are marked as optional
                    accessories = package[:items].select { |item| item[:optional] == true }

                    package_hash = {
                        id:                package[:id].to_s,
                        productCategoryId: product_category_id,
                        customerGroupId:   customer_group_id,
                        name:              package[:name],
                        alias:             package[:alias],
                        shortDescription:  package[:info],
                        description:       package[:details],
                        producer:          package[:producer],
                        stars:             package[:category],
                        image:             "https://sports-rental.info" + package[:img],
                        mainProduct:       self.convert_item_to_accessory(main_product, start_date, duration, age, shop_id, shop_lang, package[:id], nil, product_category_id, customer_group_id),
                        accessories:       self.convert_items_to_accessories(accessories, start_date, duration, age, shop_id, shop_lang, package[:id], nil, product_category_id, customer_group_id),
                        packageBackgroundColor: package[:bgColor],
                        packageTextColor:       package[:txColor],
                        productBackgroundColor: package[:pbColor],
                        productTextColor:       package[:ptColor]
                    }
                    return_packages.push package_hash
                end

            end

            return_packages
        end

        def self.products(shop_id, product_category_id, start_date, duration, customer_group_id, shop_lang)
            age = self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            categories = self.sr_products(shop_id, shop_lang)
            return_products = []
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank?

                # This lookup is kind of slow. We might need to improve later, if necessary
                next if category[:id].to_s != product_category_id

                return_products = self.convert_items_to_accessories(category[:products], start_date, duration, age, shop_id, shop_lang, nil, category[:type], product_category_id, customer_group_id)
            end

            return_products
        end

        def self.all_categories_with_packages_and_products(shop_id, start_date, duration, shop_lang)
            return_categories = []
            customer_groups = self.customer_groups(shop_id, shop_lang)
            categories = self.sr_products(shop_id, shop_lang)
            categories.each do |category|
                # Skip if there are no products in this category
                next if category[:products].blank? && category[:packages].blank?

                product_category_id = category[:id].to_s

                category_hash = {
                    id: product_category_id,
                    name: category[:name],
                    type: category[:type],
                    packages: {},
                    products: {}
                }

                unless category[:packages].blank?
                    category[:packages].each do |package|
                        customer_groups.each do |customer_group|

                            customer_group_id = customer_group[:id].to_s
                            age = customer_group[:ageFrom]

                            if category_hash[:packages][customer_group_id].blank?
                                category_hash[:packages][customer_group_id] = []
                            end

                            # nur dann dem Packages Hash hinzufügen, wenn für diese Kundengruppe ein Preis existiert
                            valid_prices = package[:prices].select { |p| age >= p[:ageFrom] && age <= p[:ageTo] }
                            next if valid_prices.empty?

                            # Main Product ist das erste nicht optionale Element in der Liste der items
                            main_product = package[:items].select { |item| item[:optional] == false }.first
                            next if main_product.blank?

                            # Accessories are all products which are marked as optional
                            accessories = package[:items].select { |item| item[:optional] == true }

                            package_hash = {
                                id:                package[:id].to_s,
                                productCategoryId: product_category_id,
                                customerGroupId:   customer_group_id,
                                name:              package[:name],
                                alias:             package[:alias],
                                shortDescription:  package[:info],
                                description:       package[:details],
                                producer:          package[:producer],
                                stars:             package[:category],
                                image:             "https://sports-rental.info" + package[:img],
                                mainProduct:       self.convert_item_to_accessory(main_product, start_date, duration, age, shop_id, shop_lang, package[:id], nil, product_category_id, customer_group_id),
                                accessories:       self.convert_items_to_accessories(accessories, start_date, duration, age, shop_id, shop_lang, package[:id], nil, product_category_id, customer_group_id),
                                packageBackgroundColor: package[:bgColor],
                                packageTextColor:       package[:txColor],
                                productBackgroundColor: package[:pbColor],
                                productTextColor:       package[:ptColor]
                            }

                            category_hash[:packages][customer_group_id].push package_hash
                        end
                    end
                end

                unless category[:products].blank?
                    customer_groups.each do |customer_group|

                        customer_group_id = customer_group[:id].to_s
                        age = customer_group[:ageFrom]

                        category_hash[:products][customer_group_id] = self.convert_items_to_accessories(category[:products], start_date, duration, age, shop_id, shop_lang, nil, category[:type], product_category_id, customer_group_id)
                    end
                end

                return_categories.push category_hash
            end # categories.each

            return_categories
        end

        def self.availability(shop_lic, date, product_id)
            self.avail(shop_lic, date, product_id)
        end

        # Helper Functions
        def self.age_for_customer_group(shop_id, shop_lang, customer_group_id)
            customer_group = self.customer_groups(shop_id, shop_lang).select{ |cg| cg[:id] == customer_group_id }.first
            # Just use the start-Age for selecting the correct price group
            customer_group[:ageFrom]
        end

        # TODO: use map..
        def self.convert_items_to_accessories(items, start_date, duration, age, shop_id, shop_lang, package_id, category, product_category_id, customer_group_id)
            accessories = []
            items.each do |item|
                valid_prices = item[:prices].select { |p| age >= p[:ageFrom] && age <= p[:ageTo] }
                next if valid_prices.empty?

                accessories.push convert_item_to_accessory(item, start_date, duration, age, shop_id, shop_lang, package_id, category, product_category_id, customer_group_id)
            end
            accessories
        end

        def self.convert_item_to_accessory(item, start_date, duration, age, shop_id, shop_lang, package_id, category, product_category_id, customer_group_id)
            config = Rails.application.config.services[:SR_API]
            image_base_url = config[:image_base_url]

            image_url = item[:img]
            if !image_url.blank? && !image_url.downcase.starts_with?('http')
                image_url = image_base_url + image_url
            end

            {
                id:                 item[:id].to_s,
                vid:                item[:vid].to_s,
                productCategoryId:  product_category_id,
                customerGroupId:    customer_group_id,
                category:           item[:type].blank? ? category : item[:type],
                name:               item[:name],
                alias:              item[:alias],
                shortDescription:   item[:info],
                description:        item[:details],
                producer:           item[:producer],
                stars:              item[:category],
                photoUrl:           image_url,
                variants:           self.convert_models_to_variants(item[:models]),
                price:              self.get_price_of_item(item[:id], start_date, duration, age, shop_id, shop_lang, package_id),
                maxCount:           item[:maxCount],
                isShortRent:        item[:isShortRent],
                stock:              item[:stock]
            }
        end

        # package_id == nil means that we want to access the package-price otherwise the product-price
        def self.get_price_of_item(id, start_date, duration, age, shop_id, shop_lang, package_id)
            item_id = package_id || id # use package id if set instad of item id
            item_prices = self.price(shop_id, shop_lang, start_date, duration, age)
            # Now find the package or product price ..
            item_price = nil
            if package_id.present?
                item_price = item_prices[:packages].select { |package_item| package_item[:id].to_i == item_id.to_i }.first
            else
                item_price = item_prices[:products].select { |product_item| product_item[:id].to_i == item_id.to_i }.first
            end

            price_hash = nil

            # Return empty hash if the API has nothing for us!
            return {} if item_price.blank?

            unless package_id.nil?
                # We have a package
                return {} if item_price.blank?
                item_price[:products].each do |product|
                    if product[:id].to_i == id.to_i
                        product[:prices].each do |price|
                            if price[:ageFrom] == age
                                price_hash = price
                                break
                            end
                        end
                        break
                    end
                end
            else
                # We have a product
                return {} if item_price.blank?
                item_price[:prices].each do |price|
                    if price[:ageFrom] == age
                        price_hash = price
                        break
                    end
                end
            end

            return {} if price_hash.nil?

            created_item = {
                id:                    price_hash[:id],
                listPrice:             price_hash[:basePrc],
                discountPercentage:    price_hash[:discntP],
                discountValue:         price_hash[:discnt].round(2),
                insurancePercentage:   price_hash[:insuranceP],
                insuranceValue:        price_hash[:insurance],
                vatPercentage:         price_hash[:taxP],
                vatValue:              price_hash[:dscTotalTax].round(2),
                vatValueWithInsurance: price_hash[:tax].round(2),
                price:                 price_hash[:dscTotal],
                priceWithInsurance:    price_hash[:value] # in case of settings.money.tax.include this is the gross price, otherwise the net price…
            }

            # Remove all nil objects. Otherwise the S|R might explode!
            # FIXME: -> NO, this object is only consumed by our frontend and payment-api <- or not!?
            created_item #.clean!
        end

        def self.convert_models_to_variants(models)
            ret_variants = []
            return ret_variants if models.nil?

            models.each do |model|
                variant_hash = { id: model[:id], vid: model[:vid], name: model[:name], alias: model[:alias] }
                ret_variants.push variant_hash
            end

            ret_variants
        end

        # class Booking
        def self.booking_data_structure(shop_lic, creation_time = nil)
            {
                bookings: [],                                           # Bookings ist die Liste mit den Buchungen der assoziierten Kunden, Kinder Partner etc
                orders: [],                                             # Liste der Posten der Buchung
                from: creation_time.present? ? creation_time : Time.now.utc.strftime('%Y-%m-%dT%H:%M:%S'),
                action: nil,
                #result: nil,
                #state: nil,
                id: 0,
                code: "",
                #frontendURL: nil, # TODO: Fill me
                client: {
                    # TODO: Fill
                    id: 0,
                    code: "",
                    firstName: "",
                    lastName: "",
                    company: "",
                    sex: "",
                    address: "",
                    #card: nil,
                    #data: nil,
                    #hotel: nil
                },
                note: "",
                shop: shop_lic,
                useInsurance: false,
                agent:  "",
                #editUntil: nil
            }
        end

        def self.set_customer_data(dest_client, source_personal_data)
            dest_client[:firstName] = source_personal_data[:firstname]
            dest_client[:lastName]  = source_personal_data[:lastname]
            dest_client[:sex]       = source_personal_data[:gender] if source_personal_data[:gender].present? # TODO: get rid of this
            dest_client[:sex]       = source_personal_data[:sex] if source_personal_data[:sex].present?
            dest_client[:company]   = source_personal_data[:company]
            dest_client[:address]   = source_personal_data[:address]
            puts "source_personal_data[:birthday]"

            # TODO: Add all data..
            dest_client[:data] = {
                id: 0,
                metric:     source_personal_data[:metric] == 'metric' ? true : false,
                size:       source_personal_data[:height].to_i,
                weight:     source_personal_data[:weight].to_i,
                ski:        source_personal_data[:skiSkillLevel],
                snobo:      source_personal_data[:snowboardPosition],
                shoesize:   source_personal_data[:shoeSize].to_f,
                shoetype:   source_personal_data[:shoeSizeSystem], # -> we retrieve all possible size/system combination as a map [mm][system][systemspecific value] but are saving only the "key" mondo
                head:       source_personal_data[:headCircumference],
                confection: source_personal_data[:clothingSize]
                #age # TODO: remove? there is no corresponding item in the sr-apis mandatory fields…
            }

            unless source_personal_data[:birthday].blank?
                dest_client[:data][:birthday] = DateTime.parse(source_personal_data[:birthday]).utc.strftime('%Y-%m-%dT%H:%M:%S')
            end
            dest_client.clean!
        end

        def self.set_customer_selected_package_variants(dest_orders, source_packages)

            source_packages.each do |package|
                dest_orders.each do |order|
                    found_coresponding_order_item = false
                    unless order[:items].blank?
                        order[:items].each do |item|
                            if item[:productId] == package[:id].to_i
                                # mainProduct
                                if package[:mainProduct][:selectedVariant].present? and package[:mainProduct][:selectedVariant][:id].present?
                                    unless item[:subitems].blank?
                                        item[:subitems].each do |subitem|
                                            if subitem[:productId] == package[:mainProduct][:id].to_i
                                                subitem[:modelId]  = package[:mainProduct][:selectedVariant][:id].to_i
                                                subitem[:modelVid] = package[:mainProduct][:selectedVariant][:vid]
                                                subitem[:model]    = package[:mainProduct][:selectedVariant][:name]
                                                break
                                            end
                                        end
                                    end
                                end

                                # accessories
                                package[:accessories].each do |accessorie|
                                    if accessorie[:selectedVariant].present? and accessorie[:selectedVariant][:id].present?
                                        unless item[:subitems].blank?
                                            item[:subitems].each do |subitem|
                                                if subitem[:productId] == accessorie[:id].to_i
                                                    subitem[:modelId]  = accessorie[:selectedVariant][:id].to_i
                                                    subitem[:modelVid] = accessorie[:selectedVariant][:vid]
                                                    subitem[:model]    = accessorie[:selectedVariant][:name]
                                                    break
                                                end
                                            end
                                        end
                                    end
                                end

                                found_coresponding_order_item = true
                                break
                            end
                        end
                    end
                    break if found_coresponding_order_item
                end
            end
        end

        def self.set_customer_selected_product_variants(dest_orders, source_products)
            source_products.each do |product|
                if product[:selectedVariant].present? and product[:selectedVariant][:id].present?
                    dest_orders.each do |order|
                        found_coresponding_order_item = false
                        unless order[:items].blank?
                            order[:items].each do |item|
                                if item[:productId] == product[:id].to_i
                                    item[:modelId]  = product[:selectedVariant][:id].to_i
                                    item[:modelVid] = product[:selectedVariant][:vid]
                                    item[:model]    = product[:selectedVariant][:name]
                                    found_coresponding_order_item = true
                                    break
                                end
                            end
                        end
                        break if found_coresponding_order_item
                    end
                end
            end
        end

        def self.create_booking(booking, action = ACTION_CREATE)
            puts "create_booking with ACTION: #{action}"

            # First read an already created booking context from the API and merge its data..
            booking_data = self.get_booking(booking.booking_id)
            unless booking_data.nil?
                booking_hash = booking_data
                fail "No Booking Context available with booking id #{booking.booking_id}" if booking_hash.empty?
            end

            shop_lic = self.get_mapping[booking.shop_id.downcase]
            fail "shop_lic was not found" if shop_lic.nil?

            booking_data = self.booking_data_structure(shop_lic, booking.creation_time)
            booking_data.merge!(booking_hash)
            booking_data[:action] = action
            booking_data[:note] = booking.context[:customerNotice] # customer notice
            if booking.context[:arrivalDate].present?
                booking_data[:arival] = booking.context[:arrivalDate] # TODO: (sic!)
            end

            # Clear old booking data before rewrite them
            booking_data[:orders] = []

            # Create booking context and ..
            first_customer = true
            booking.context[:customers].each do |customer|
                orders = self.create_orders(customer, booking)

                # This API seems to expect that one persion is the main customer who
                # does multiple bookings for his relatives or friends.
                if first_customer
                    booking_data[:orders] = orders
                    booking_data[:useInsurance] = customer[:useInsurance]

                    # After adding the main customer, we need to switch to the relatives
                    booking_data[:bookings] = []
                    first_customer = false
                else
                    booking_data[:bookings].push self.create_sub_booking(customer, booking, shop_lic, action)
                end
            end

            booking_data
        end

        # class SubBooking
        def self.create_sub_booking(customer, booking, shop_lic, action)
            # Create SubBooking-Object
            {
                orders: self.create_orders(customer, booking),
                bookingId: booking.booking_id.to_i,
                action: action,
                # result
                # state
                id: 0,
                # code
                client:{
                    # TODO: Fill
                    id: 0,
                    code: "",
                    firstName: "",
                    lastName: "",
                    sex: "",
                    company: "",
                    address: nil,
                    card: nil,
                    data: nil,
                    hotel: nil
                },
                note: "",
                shop: shop_lic,
                useInsurance: customer[:useInsurance]
            }
        end

        def self.booking_date(booking, item, product)
            if (booking[:context][:isShortRent])
                min_time = 24
                max_time = 0
                product[:selectedLendingBlocks].keys.each do |time|
                    start_time = product[:selectedLendingBlocks][time][:start]
                    end_time = product[:selectedLendingBlocks][time][:end]
                    if (min_time > start_time)
                        min_time = start_time
                    end
                    if (max_time < end_time)
                        max_time = end_time
                    end
                end
                from_hour = min_time.floor
                from_min = ((min_time - from_hour) * 60).round
                until_hour = max_time.floor
                until_min = ((max_time - until_hour) * 60).round
                from_time = "#{booking[:context][:startDate]}T#{from_hour.to_s.rjust(2, '0')}:#{from_min.to_s.rjust(2, '0')}:00"
                until_time = "#{booking[:context][:endDate]}T#{until_hour.to_s.rjust(2, '0')}:#{until_min.to_s.rjust(2, '0')}:00"
                {
                    :from => from_time,
                    :until => until_time
                }
            else
                {
                    :from  => booking[:context][:startDate],
                    :until => booking[:context][:endDate]
                }
            end
        end

        def self.order_header(booking, item, product)
            booking_time = self.booking_date(booking, item, product)
            {
                :id    => 0,
                :Code  => "",
                :from  => booking_time[:from],
                :until => booking_time[:until],
                :items => [ item ],
                :isSeason => booking[:context][:useSeasonalRental],
                :isShortRent => booking[:context][:isShortRent]
            }
        end

        def self.create_orders(customer, booking)
            # Handle Packages
            orders = []
            customer[:packages].each do |package|
                package[:mainProduct][:count].times do
                    # Package data
                    item = self.create_item(package)

                    sub_items = []

                    # The main item is a sub item as well ..
                    sub_items.push self.create_subitem(package[:mainProduct])

                    # Move all package optionals as subitems!
                    package[:accessories].each do |accessory|
                        sub_items.push self.create_subitem(accessory)
                    end

                    item[:subitems] = sub_items
                    orders.push self.order_header(booking, item, package[:mainProduct])
                end
            end
            # Handle Products which are not in packages
            customer[:products].each do |product|
                product[:count].times do
                    item = self.create_item(product)
                    orders.push self.order_header(booking, item, product)
                end
            end

            orders
        end

        def self.create_item(product)
            created_item = {
                subitems:   [],
                orderId:    0,
                id:         0,
                #discnt:     product[:price][:discountValue],
                #discntP:    product[:price][:discountPercentage],
                #insure:     product[:price][:insuranceValue],
                #insureP:    product[:price][:insurancePercentage],
                #price:      product[:price][:listPrice],
                #total:      product[:price][:grossPrice],
                productId:  product[:id].to_i,
                productVid: nil,
                product:    product[:name],
                alias:      product[:alias],
                modelId:    (product[:selectedVariant].blank? ? 0 : product[:selectedVariant][:id].to_i),
                modelVid:   "",
                model:      (product[:selectedVariant].blank? ? "" : product[:selectedVariant][:name]),
                priceId:    product[:price].present? ? product[:price][:id] : nil
            }

            created_item.clean!
        end

        def self.create_subitem(product)
            created_item = {
                itemId:     0,
                id:         0,
                alias:      product[:alias],
                model:      (product[:selectedVariant].blank? ? "" : product[:selectedVariant][:name]),
                modelId:    (product[:selectedVariant].blank? ? 0 : product[:selectedVariant][:id].to_i),
                modelVid:   (product[:selectedVariant].blank? ? "" : product[:selectedVariant][:vid]),
                priceId:    product[:price][:id],
                product:    product[:name],
                productId:  product[:id].to_i,
                productVid: product[:vid],
                #discnt:     product[:price][:discountValue],
                #discntP:    product[:price][:discountPercentage],
                #insure:     product[:price][:insuranceValue],
                #insureP:    product[:price][:insurancePercentage],
                #price:      product[:price][:listPrice],
                #total:      product[:price][:grossPrice]
            }

            created_item.clean!
        end

        # Low-Level requests from SR Backend
        private

        def self.service_connection
            config = Rails.application.config.services[:SR_API]
            service_url = config[:base_url]
            Faraday.new(:url => service_url) do |f|
                f.use Faraday::AddSrTokens
                f.basic_auth(config[:login], config[:password])
                f.response :logger                                                      # log requests to STDOUT
                f.use Faraday::HttpCache, store: Rails.cache, logger: Rails.logger
                f.use FaradayMiddleware::ParseJson, content_type: 'application/json'
                f.adapter Faraday.default_adapter                                      # make requests with Net::HTTP
            end
        end

        def self.api_version
            response = service_connection.get("/srapi/")
            fail if response.status != 200
            response.body[:version]
        end

        def self.login
            response = service_connection.get("/srapi/login")
            fail if response.status != 200
            response.body
        end

        def self.settings(shop_id, shop_lang = "de")
            response = service_connection.get("/srapi/settings/#{URI.encode(shop_id)}?lang=#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.sr_products(shop_id, shop_lang)
            response = service_connection.get("/srapi/product/#{URI.encode(shop_id)}/#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.client(shop_id, shop_lang)
            response = service_connection.get("/srapi/client/#{URI.encode(shop_id)}?lang=#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.shop(shop_id, shop_lang)
            response = service_connection.get("/srapi/shop/#{URI.encode(shop_id)}?lang=#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.ticket
            response = service_connection.get("/srapi/ticket")
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.price(shop_id, shop_lang, start, duration, age)
            response = service_connection.get("/srapi/price/#{URI.encode(shop_id)}/#{shop_lang_to_two_letters(shop_lang)}/#{start}/#{duration}/#{age}")
            fail if response.status != 200 && response.status != 204
            unless response.body.empty?
                response.body.deep_symbolize_keys
            else
                return nil
            end
        end

        def self.price_info(shop_id, shop_lang)
            response = service_connection.get("/srapi/priceinfo/#{URI.encode(shop_id)}?lang=#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.avail(shop_lic, date, product_id)
            response = service_connection.get("/srapi/avail/#{shop_lic}/#{date}/#{product_id}")
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.get_booking(booking_id)
            response = service_connection.get("/srapi/booking/#{booking_id}" )
            fail if response.status != 200 && response.status != 204
            unless response.body.empty?
                response.body.deep_symbolize_keys
            else
                return nil
            end
        end

        def self.accept_booking(booking_id)
            response = service_connection.get("/srapi/booking/accept/#{booking_id}" )
            fail if response.status != 200 && response.status != 204
            unless response.body.empty?
                response.body.deep_symbolize_keys
            else
                return nil
            end
        end

        def self.post_booking(numeric_shop_id, booking_data, shop_lang = "de")
            Rails.logger.debug "Posting Booking-Object to SR-API (/srapi/booking/#{numeric_shop_id}): #{booking_data.to_json} - lang: #{shop_lang}"
            response = service_connection.post do |req|
                req.url("/srapi/booking/#{numeric_shop_id}?lang=#{shop_lang_to_two_letters(shop_lang)}")
                req.headers['Content-Type'] =   'application/json'
                req.body = booking_data.to_json
            end

            Rails.logger.debug "Post-Booking-Response: #{response.body.to_json} - Status: #{response.status}"
            fail if response.status != 200

            response_data = response.body.deep_symbolize_keys
            fail "Something went wrong with post_booking. Result:#{response_data[:result]}" if response_data[:result] > 2

            response_data
        end

        def self.get_mapping
            response = service_connection.get( "/srapi/Mapping" )
            fail if response.status != 200
            response.body
        end

        def self.metrics(shop_id, metric_type, shop_lang)
            # Curr. we do not need sho_id here, but maybe in future?
            response = service_connection.get("/srapi/metrics/#{metric_type}?lang=#{shop_lang_to_two_letters(shop_lang)}" )
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.availableLanguages()
            response = service_connection.get("/srapi/Local")
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.i18n(shop_lang, shop_lic = nil)
            url = "/srapi/Local/#{shop_lang_to_two_letters(shop_lang)}"
            if shop_lic.present?
                url = url + "/#{shop_lic}"
            end
            puts "calling #{url} …"
            response = service_connection.get(url)
            fail if response.status != 200
            response.body.deep_symbolize_keys
        end

        def self.shop_lang_to_two_letters(shop_lang)
            shop_lang.split("-").first
        end

    end
end
